@@ -4,12 +4,12 @@ use nom::{
     branch::alt,
     bytes::complete::tag,
     character::complete::{alpha1, char, not_line_ending, satisfy, space0, space1},
-    combinator::opt,
-    error::{context, convert_error, VerboseError},
+    combinator::{opt, verify},
+    error::{context, VerboseError},
     multi::many1,
     number::complete::float,
     sequence::{delimited, tuple},
-    Err as NomErr, IResult,
+    IResult,
 };
 
 extern crate nom;
@@ -18,6 +18,9 @@ extern crate nom;
 #[macro_use]
 extern crate serde;
 
+pub mod rich_text;
+mod unit;
+
 type Res<T, U> = IResult<T, U, VerboseError<T>>;
 
 #[cfg_attr(feature = "serde-derive", derive(Serialize, Deserialize))]
@@ -64,7 +67,7 @@ pub struct Ingredient {
 impl TryFrom<&str> for Ingredient {
     type Error = String;
     fn try_from(value: &str) -> Result<Ingredient, Self::Error> {
-        from_str(value, true)
+        Ok(from_str(value))
     }
 }
 
@@ -78,7 +81,7 @@ impl fmt::Display for Ingredient {
             None => "".to_string(),
         };
         let amount_list = match amounts.len() {
-            0 => "n/a".to_string(),
+            0 => "n/a ".to_string(),
             _ => format!("{} ", amounts.join(" / ")),
         };
         let name = self.name.clone();
@@ -88,51 +91,28 @@ impl fmt::Display for Ingredient {
 /// wrapper for [parse_ingredient]
 /// ```
 /// use ingredient::{from_str};
-/// assert_eq!(from_str("one whole egg", true).unwrap().to_string(),"1 whole egg");
+/// assert_eq!(from_str("one whole egg").to_string(),"1 whole egg");
 /// ```
-pub fn from_str(input: &str, verbose_error: bool) -> Result<Ingredient, String> {
-    return match parse_ingredient(input) {
-        Ok(r) => Ok(r.1),
-        Err(e) => {
-            let msg = match e {
-                NomErr::Error(e) => {
-                    if verbose_error {
-                        convert_error(input, e)
-                    } else {
-                        format!("{}", e)
-                    }
-                }
-                _ => format!("{}", e),
-            };
-            return Err(format!("failed to parse '{}': {}", input, msg));
-        }
-    };
+pub fn from_str(input: &str) -> Ingredient {
+    //todo: add back error handling? can't get this to ever fail since parser is pretty flexible
+    parse_ingredient(input).unwrap().1
 }
+
 /// Parses one or two amounts, e.g. `12 grams` or `120 grams / 1 cup`. Used by [parse_ingredient].
 /// ```
 /// use ingredient::{parse_amount,Amount};
 /// assert_eq!(
-///    parse_amount("120 grams").unwrap(),
+///    parse_amount("120 grams"),
 ///    vec![Amount::new("grams",120.0)]
 ///  );
 /// assert_eq!(
-///    parse_amount("120 grams / 1 cup").unwrap(),
+///    parse_amount("120 grams / 1 cup"),
 ///    vec![Amount::new("grams",120.0),Amount::new("cup", 1.0)]
 ///  );
 /// ```
-pub fn parse_amount(input: &str) -> Result<Vec<Amount>, String> {
-    return match parse_n_amount(input) {
-        Ok(r) => Ok(r.1),
-        Err(e) => {
-            let msg = match e {
-                NomErr::Error(e) => {
-                    format!("{}", e)
-                }
-                _ => format!("{}", e),
-            };
-            return Err(format!("failed to parse '{}': {}", input, msg));
-        }
-    };
+pub fn parse_amount(input: &str) -> Vec<Amount> {
+    // todo: also can't get this one to fail either
+    many_amount(input).unwrap().1
 }
 
 /// Parse an ingredient line item, such as `120 grams / 1 cup whole wheat flour, sifted lightly`.
@@ -180,7 +160,7 @@ pub fn parse_ingredient(input: &str) -> Res<&str, Ingredient> {
     context(
         "ing",
         tuple((
-            opt(parse_n_amount),
+            opt(many_amount),
             space0,           // space between amount(s) and name
             opt(many1(text)), // name, can be multiple words
             opt(amt_parens),  // can have some more amounts in parens after the name
@@ -192,10 +172,12 @@ pub fn parse_ingredient(input: &str) -> Res<&str, Ingredient> {
         let (amounts, _, name_chunks, amounts2, _, modifier_chunks) = res;
         let m = modifier_chunks;
 
-        let mut name = match name_chunks {
-            Some(n) => n.join("").trim_matches(' ').to_string(),
-            None => "".to_string(),
-        };
+        let name = name_chunks
+            .unwrap_or(vec![])
+            .join("")
+            .trim_matches(' ')
+            .to_string();
+
         let mut amounts = match amounts {
             Some(a) => a,
             None => vec![],
@@ -205,12 +187,6 @@ pub fn parse_ingredient(input: &str) -> Res<&str, Ingredient> {
             None => amounts,
         };
 
-        // if we have a unit but no name, e.g. `1 egg`,
-        // the unit is the name, so normalize it to `1 whole egg`,
-        if name.len() == 0 && amounts.len() == 1 {
-            name = amounts[0].unit.clone();
-            amounts[0].unit = "whole".to_string();
-        };
         (
             next_input,
             Ingredient {
@@ -226,7 +202,7 @@ pub fn parse_ingredient(input: &str) -> Res<&str, Ingredient> {
 }
 
 fn text(input: &str) -> Res<&str, &str> {
-    alt((alpha1, space1, tag("-")))(input)
+    alt((alpha1, space1, tag("-"), tag("'")))(input)
 }
 
 // parses 2 amounts, seperated by ; or /
@@ -250,7 +226,12 @@ fn num_or_range(input: &str) -> Res<&str, (f32, Option<f32>)> {
         tuple((
             num,
             opt(tuple(
-                (space0, tag("-"), space0, num), // care about u.3
+                (
+                    space0,
+                    alt((tag("-"), tag("–"))), // second dash is an unusual variant
+                    space0,
+                    num,
+                ), // care about u.3
             )),
         )),
     )(input)
@@ -263,6 +244,10 @@ fn num_or_range(input: &str) -> Res<&str, (f32, Option<f32>)> {
         (next_input, (val, upper))
     })
 }
+
+fn unit(input: &str) -> Res<&str, &str> {
+    context("unit", verify(alpha1, |s: &str| unit::is_valid(s)))(input)
+}
 // parses a single amount
 fn amount1(input: &str) -> Res<&str, Vec<Amount>> {
     context(
@@ -272,17 +257,17 @@ fn amount1(input: &str) -> Res<&str, Vec<Amount>> {
                 opt(tag("about ")), // todo: add flag for estimates
                 num_or_range,       // value
                 space0,
-                alpha1, // unit
+                opt(unit), // unit
+                opt(tag(".")),
             ), // 1 gram
         ),
     )(input)
     .map(|(next_input, res)| {
-        let (_, value, _, unit) = res;
-
+        let (_prefix, value, _space, unit, _period) = res;
         (
             next_input,
             vec![Amount {
-                unit: unit.to_string(),
+                unit: unit.unwrap_or("whole").to_string().to_lowercase(),
                 value: value.0,
                 upper_value: value.1,
             }],
@@ -291,7 +276,7 @@ fn amount1(input: &str) -> Res<&str, Vec<Amount>> {
 }
 
 // parses one or two amounts, e.g. `12 grams` or `120 grams / 1 cup`
-fn parse_n_amount(input: &str) -> Res<&str, Vec<Amount>> {
+fn many_amount(input: &str) -> Res<&str, Vec<Amount>> {
     context(
         "amount",
         alt((
@@ -304,13 +289,12 @@ fn parse_n_amount(input: &str) -> Res<&str, Vec<Amount>> {
 }
 
 fn amt_parens(input: &str) -> Res<&str, Vec<Amount>> {
-    context(
-        "amt_parens",
-        delimited(char('('), parse_n_amount, char(')')),
-    )(input)
+    context("amt_parens", delimited(char('('), many_amount, char(')')))(input)
 }
 
-fn v_frac_to_num(input: &char) -> Result<f32, String> {
+fn v_frac_to_num(input: char) -> Result<f32, String> {
+    // two ranges for unicode fractions
+    // https://www.compart.com/en/unicode/search?q=vulgar+fraction#characters
     let (n, d): (i32, i32) = match input {
         '¾' => (3, 4),
         '⅛' => (1, 8),
@@ -321,25 +305,14 @@ fn v_frac_to_num(input: &char) -> Result<f32, String> {
     };
     return Ok(n as f32 / d as f32);
 }
-// two ranges for unicode fractions
-// https://www.compart.com/en/unicode/search?q=vulgar+fraction#characters
+
 fn is_frac_char(c: char) -> bool {
-    match c {
-        '¼'..='¾' => true,
-        '⅐'..='⅞' => true,
-        _ => false,
-    }
+    v_frac_to_num(c).is_ok()
 }
 /// parses unicode vulgar fractions
 fn v_fraction(input: &str) -> Res<&str, f32> {
-    context("v_fraction", satisfy(is_frac_char))(input).map(|(next_input, res)| {
-        let num = match v_frac_to_num(&res) {
-            Ok(x) => x,
-            _ => 0.0,
-        };
-
-        (next_input, num)
-    })
+    context("v_fraction", satisfy(is_frac_char))(input)
+        .map(|(next_input, res)| (next_input, v_frac_to_num(res).unwrap()))
 }
 
 fn n_fraction(input: &str) -> Res<&str, f32> {
@@ -382,6 +355,7 @@ fn fraction_number(input: &str) -> Res<&str, f32> {
 #[cfg(test)]
 mod tests {
     use nom::error::{ErrorKind, VerboseErrorKind};
+    use nom::Err as NomErr;
 
     use super::*;
     #[test]
@@ -391,7 +365,6 @@ mod tests {
         assert_eq!(fraction_number("1⅓"), Ok(("", 1.3333334)));
         assert_eq!(fraction_number("¼"), Ok(("", 0.25)));
         assert_eq!(fraction_number("1/4"), Ok(("", 0.25)));
-        assert_eq!(fraction_number("⅐"), Ok(("", 0.0))); // unkown are dropped
         assert_eq!(
             fraction_number("1"),
             Err(NomErr::Error(VerboseError {
@@ -406,21 +379,18 @@ mod tests {
     }
     #[test]
     fn test_v_fraction() {
-        assert_eq!(v_frac_to_num(&'⅛'), Ok(0.125));
-        assert_eq!(v_frac_to_num(&'¼'), Ok(0.25));
-        assert_eq!(v_frac_to_num(&'½'), Ok(0.5));
+        assert_eq!(v_frac_to_num('⅛'), Ok(0.125));
+        assert_eq!(v_frac_to_num('¼'), Ok(0.25));
+        assert_eq!(v_frac_to_num('½'), Ok(0.5));
     }
 
     #[test]
     fn test_amount_range() {
         assert_eq!(
-            parse_amount("2¼-2.5 cups").unwrap(),
+            parse_amount("2¼-2.5 cups"),
             vec![Amount::new_with_upper("cups", 2.25, 2.5)]
         );
-        assert_eq!(
-            parse_amount("2¼-2.5 cups").unwrap(),
-            parse_amount("2 ¼ - 2.5 cups").unwrap()
-        );
+        assert_eq!(parse_amount("2¼-2.5 cups"), parse_amount("2 ¼ - 2.5 cups"));
         assert_eq!(
             Ingredient::try_from("1-2 cups flour"),
             Ok(Ingredient {
@@ -440,6 +410,10 @@ mod tests {
                 modifier: None,
             })
         );
+    }
+
+    #[test]
+    fn test_ingredient_parse_no_amounts() {
         assert_eq!(
             parse_ingredient("egg"),
             Ok((
@@ -451,6 +425,9 @@ mod tests {
                 }
             ))
         );
+    }
+    #[test]
+    fn test_ingredient_parse_no_unit() {
         assert_eq!(
             parse_ingredient("1 egg"),
             Ok((
@@ -466,20 +443,76 @@ mod tests {
                 }
             ))
         );
+    }
+    #[test]
+    fn test_ingredient_parse_no_unit_multi_name() {
         assert_eq!(
-            format!("res: {}", from_str("12 cups flour", false).unwrap()),
+            parse_ingredient("1 cinnamon stick"),
+            Ok((
+                "",
+                Ingredient {
+                    name: "cinnamon stick".to_string(),
+                    amounts: vec![Amount {
+                        unit: "whole".to_string(),
+                        value: 1.0,
+                        upper_value: None,
+                    }],
+                    modifier: None,
+                }
+            ))
+        );
+        assert_eq!(
+            parse_ingredient("1 cinnamon stick"),
+            parse_ingredient("1 whole cinnamon stick"),
+        );
+    }
+    #[test]
+    fn test_ingredient_parse_no_unit_multi_name_adj() {
+        assert_eq!(
+            parse_ingredient("1 cinnamon stick, crushed"),
+            Ok((
+                "",
+                Ingredient {
+                    name: "cinnamon stick".to_string(),
+                    amounts: vec![Amount {
+                        unit: "whole".to_string(),
+                        value: 1.0,
+                        upper_value: None,
+                    }],
+                    modifier: Some("crushed".to_string()),
+                }
+            ))
+        );
+    }
+    #[test]
+    fn test_stringy() {
+        assert_eq!(
+            format!("res: {}", from_str("12 cups flour")),
             "res: 12 cups flour"
         );
+        assert_eq!(from_str("one whole egg").to_string(), "1 whole egg");
+    }
+    #[test]
+    fn test_with_parens() {
         assert_eq!(
-            from_str("one whole egg", true).unwrap().to_string(),
-            "1 whole egg"
-        );
-        assert_eq!(
-            from_str("1 cup (125.5 grams) AP flour, sifted", false)
-                .unwrap()
-                .to_string(),
+            from_str("1 cup (125.5 grams) AP flour, sifted").to_string(),
             "1 cup / 125.5 grams AP flour, sifted"
         );
+    }
+    #[test]
+    fn test_no_ingredient_amounts() {
+        assert_eq!(
+            Ingredient {
+                name: "apples".to_string(),
+                amounts: vec![],
+                modifier: None,
+            }
+            .to_string(),
+            "n/a apples"
+        );
+    }
+    #[test]
+    fn test_ingredient_parse_multi() {
         assert_eq!(
             parse_ingredient("12 cups all purpose flour, lightly sifted"),
             Ok((
@@ -545,4 +578,51 @@ mod tests {
             parse_ingredient("1 1/2 cups / 192 grams all-purpose flour")
         );
     }
+    #[test]
+    fn test_weird_chars() {
+        assert_eq!(
+            parse_ingredient("100g confectioner's sugar, sifted"),
+            Ok((
+                "",
+                Ingredient {
+                    name: "confectioner's sugar".to_string(),
+                    amounts: vec![Amount::new("g", 100.0),],
+                    modifier: Some("sifted".to_string())
+                }
+            ))
+        );
+    }
+    #[test]
+    fn test_unit_period_mixed_case() {
+        assert_eq!(
+            parse_ingredient("1 Tbsp. flour"),
+            parse_ingredient("1 tbsp flour"),
+        );
+    }
+    #[test]
+    fn test_parse_ingredient_cloves() {
+        assert_eq!(
+            parse_ingredient("1 clove garlic, grated"),
+            Ok((
+                "",
+                Ingredient {
+                    name: "garlic".to_string(),
+                    amounts: vec![Amount::new("clove", 1.0),],
+                    modifier: Some("grated".to_string())
+                }
+            ))
+        );
+        //todo: doesn't work
+        // assert_eq!(
+        //     parse_ingredient("1 clove, grated"),
+        //     Ok((
+        //         "",
+        //         Ingredient {
+        //             name: "clove".to_string(),
+        //             amounts: vec![Amount::new("whole", 1.0),],
+        //             modifier: Some("grated".to_string())
+        //         }
+        //     ))
+        // );
+    }
 }
